{
********************
* unit hawkVoice.pas
* Version 0.91
* 31/10/2004
*
* HawkDI is (c) HawkSoftware (http://www.hawksoft.com)
* Initial convertion by Matheus Degiovani (matheus@tilt.net)
*******************
}


//from hvdi.h and hvdint.h
unit hawkVoice;

interface

uses
     windows, classes, sysUtils;

{$DEFINE AUTOLOAD} //loads dll with unit initialization

{$IFNDEF HVDI_H}
{$DEFINE HVDI_H}
{$ENDIF}

const
     BF_ROUNDS	= 16; //from blowfish.h
     HVDI_DLL = 'hvdi.dll';

type
     PByte = ^Byte;
     PSmallInt = ^SmallInt;
     PInteger = ^Integer;
     PPSmallInt = ^PSmallInt;

     //from blowfish.h
     BF_KEY = record
     	P: array[0..BF_ROUNDS+1] of cardinal; //unsigned long
	     S: array[0..4*256-1] of cardinal;
	end;

     //from hcrypt.c
     hcrypt_key = record
          bf: BF_KEY;
          iv: array[0..7] of byte;
          digest: array[0..15] of byte;
          n: integer;
     end;
     Phcrypt_key = ^hcrypt_key;

     //from rate.c
     hvdi_rate = record
          lcmrate: integer;		    { least common multiple of rates }
          inskip, outskip: integer;    { LCM increments for I & O rates }
          total: integer;
          intot, outtot: integer;      { total samples in terms of LCM rate }
          lastsamp: integer;
     end;
     Phvdi_rate = ^hvdi_rate;

     hvdi_state = record
          gsm_lpt: integer;
          celp_codebook: integer;
          celp_fast_gain: integer;
          sequence: integer;
          autoVOX: integer;
          VOXlevel: integer;
          VOXspeed: integer;
          comfortnoise: integer;
          noiselevel: integer;
     end;
     hvdi_state_t = ^hvdi_state;

     hvdi_vox = record
          rate: integer;           { HVDI_VOX_FAST, HVDI_VOX_MEDIUM, or HVDI_VOX_SLOW }
          noisethreshold: integer; { The actual threshold used by hvdiVOX }
          samplecount: integer;    { init to 0; used internally by hvdiVOX }
     end;
     hvdi_vox_t = ^hvdi_vox;
     Phvdi_vox = ^hvdi_vox;

     hvdi_enc_state = record
          codec: byte;      { the codec used with the last packet }
          sequence: word;   { the sequence number of the last packet }
          state: pointer;     { the codec state }
          vox: hvdi_vox_t;  { the VOX structure for auto VOX }
     end;
     Phvdi_enc_state = ^hvdi_enc_state;

     hvdi_dec_state = record
          codec: byte;      { the codec used with the last packet }
          sequence: word;   { the sequence number of the last packet }
          state: pointer;     { the codec state }
     end;
     Phvdi_dec_state = ^hvdi_dec_state;

     hvdi_agc = record
          sample_max: cardinal;
          counter: integer;
          igain: integer; //long?
          ipeak: integer;
          silence_counter: integer;
     end;
     Phvdi_agc = ^hvdi_agc;


const
     HVDI_MAJOR_VERSION = 0;
     HVDI_MINOR_VERSION = 91;
     HVDI_VERSION_STRING = 'HVDI 0.91 beta';

{$IFNDEF NL_INVALID}
     NL_INVALID = -1;
     NL_FALSE   = 0;
     NL_TRUE    = 1;
{$ENDIF}

//FIXME: Not yet translated from original
{$IFDEF HVDI_OLD_API}
     hvdiCreateEncoderState hvdiNewEncState;
     hvdiCreateDecoderState  hvdiNewDecState
     hvdiFreeEncoderState    hvdiDeleteEncState
     hvdiFreeDecoderState    hvdiDeleteDecState
     hvdiSetCodec(x, y)      hvdiEncStateSetCodec(y, x)
     hvdiIsVoicePacket       hvdiPacketIsVoice
     hvdiDecodePacket        hvdiPacketDecode
     hvdiEncodePacket(a, b, c, d, e, f) hvdiPacketEncode(c, b, a, d, e, f)
{$ENDIF}

const
     //The basic codecs, from hawkvoice.h
     HV_2_4K_CODEC       = $0001;  { LPC-10 2.4 Kbps codec }
     HV_4_8K_CODEC       = $0002;  { LPC 4.8 Kbps codec }
     HV_13_2K_CODEC      = $0003;  { GSM 13.2 Kbps codec }
     HV_32K_CODEC        = $0004;  { Intel/DVI ADPCM 32 Kbps codec }
     HV_64K_CODEC        = $0005;  { G.711 u-law 64 Kbps codec }
     HV_1_4K_CODEC       = $0006;  { OpenLPC 1.4 Kbps codec }
     HV_1_8K_CODEC       = $0007;  { OpenLPC 1.8 Kbps codec }
     HV_4_5K_CODEC       = $0008;  { CELP 4.5 Kbps codec }
     HV_3_0K_CODEC       = $0009;  { CELP 3.0 Kbps codec }
     HV_2_3K_CODEC       = $000a;  { CELP 2.3 Kbps codec }
     HV_VBR_2_4K_CODEC   = $000b;  { Variable Bit Rate LPC-10 2.4 Kbps max codec}

     HV_SILENCE_CODEC    = $001f;  { Silence codec, used internally }

{ Alternate codec names }
     HV_LPC10_CODEC      = HV_2_4K_CODEC;
     HV_LPC_CODEC        = HV_4_8K_CODEC;
     HV_GSM_CODEC        = HV_13_2K_CODEC;
     HV_ADPCM_32_CODEC   = HV_32K_CODEC;
     HV_PCM_64_CODEC     = HV_64K_CODEC;
     HV_G_711_CODEC      = HV_64K_CODEC;
     HV_ULAW_CODEC       = HV_64K_CODEC;
     HV_LPC_1_4_CODEC    = HV_1_4K_CODEC;
     HV_LPC_1_8_CODEC    = HV_1_8K_CODEC;
     HV_CELP_4_5_CODEC   = HV_4_5K_CODEC;
     HV_CELP_3_0_CODEC   = HV_3_0K_CODEC;
     HV_CELP_2_3_CODEC   = HV_2_3K_CODEC;
     HV_VBR_LPC10_CODEC  = HV_VBR_2_4K_CODEC;

{ VOX options }
{ How many samples of silence to wait after voice stops. }
{ You can use any value, these are just for reference. }
     HVDI_VOX_FAST       = 4000;    { 1/2 second }
     HVDI_VOX_MEDIUM     = 8000;    { 1 second }
     HVDI_VOX_SLOW      = 12000;    { 1 1/2 seconds }

{ hvdiHint options}
     HVDI_NORMAL         = $0001;  { Normal encoding/decoding speed, best quality, arg ignored }
     HVDI_FAST           = $0002;  { Faster encoding/decoding, some loss of quality
                                       with some codecs, arg ignored }
     HVDI_FASTEST        = $0003;  { Fastest possible encoding/decoding, more loss of quality,
                                       arg ignored }
     HVDI_CELP_CODEBOOK  = $0004;  { Directly change the CELP encoding codebook length, arg
                                       valid range 32 to 256 }
     HVDI_SEQUENCE       = $0005;  { Determines if the sequence number is sent in the packet.
                                       To disable, arg = 0, to enable (default), arg != 0 }
     HVDI_AUTO_VOX       = $0006;  { Enables automatic VOX processing inside hvdiPacketEncode.
                                       Default VOX setting is 300 unless first set by HVD_VOX_LEVEL }
     HVDI_VOX_LEVEL      = $0007;  { Sets the threshhold level when using HVDI_AUTO_VOX, arg
                                       valid range 0 to 1000, default 300 }
     HVDI_VOX_SPEED      = $0008;  { Sets the VOX speed, default HVDI_VOX_FAST }
     HVDI_COMFORT_NOISE  = $0009;  { Enables sending silence packets by the encoder, and creates
                                       white noise in the decoder }
     HVDI_NOISE_LEVEL    = $000a;  { Sets the decoder comfort noise level, 0 to 1000, default 100 }

var
     codecNames: array[1..11] of string;

var
     hvdiNewEncState: function (): Phvdi_enc_state; stdcall;

     hvdiNewDecState: function (): Phvdi_dec_state; stdcall;

     hvdiDeleteEncState: procedure (state: Phvdi_enc_state); stdcall;

     hvdiDeleteDecState: procedure (state: Phvdi_dec_state); stdcall;

     hvdiEncStateSetCodec: function (state: Phvdi_enc_state; codec: byte): integer; stdcall; //stdcall;

     hvdiDecStateGetCodec: function (state: Phvdi_dec_state): byte; stdcall;

     hvdiPacketIsVoice: function (packet: PByte; length: integer): integer; stdcall;

     hvdiPacketDecode: function (packet: PByte; paclen: integer; buffer: PSmallInt;
                                 buflen: integer; key: Phcrypt_key;
                                 state: Phvdi_dec_state): integer; stdcall;

     hvdiPacketEncode: function (packet: PByte; buflen: integer; buffer: PSmallInt;
                                 paclen: integer; key: Phcrypt_key;
                                 state: Phvdi_enc_state): integer; stdcall;

     hvdiNewVOX: function (vox_speed: integer; noisethreshold: integer): Phvdi_vox; stdcall;

     hvdiVOX: function (vox: Phvdi_vox; buffer: PSmallInt; buflen: integer): integer; stdcall;

     hvdiDeleteVOX: procedure(vox: Phvdi_vox); stdcall;

     hvdiNewRate: function (inrate: integer; outrate: integer): Phvdi_rate; stdcall;

     hvdiRateFlow: procedure (rate: Phvdi_rate; inbuf: PSmallInt; outbuf: PSmallInt;
                              inlen: PInteger; outlen: PInteger); stdcall;

     hvdiDeleteRate: procedure (rate: Phvdi_rate); stdcall;

     hvdiNewAGC: function (level: single): Phvdi_agc; stdcall;

     hvdiAGC: procedure (agc: Phvdi_agc; buffer: PSmallInt; len: integer); stdcall;

     hvdiDeleteAGC: procedure (agc: Phvdi_agc); stdcall;

     hvdiMix: procedure (outbuf: PSmallInt; inbuf: PPSmallInt; number: integer;
                         inlen: integer); stdcall;

     hvdiHint: procedure (name: integer; arg: integer); stdcall;

function initHawkDI: boolean;
procedure FreeHawkDI;
function hvdiHighestFramesize: integer;

implementation

var
  dllHandle:THandle=0;

function getWindowsErrorText: string;
var
errorCode: cardinal;
p: PChar;
begin
     errorCode:= GetLastError;
     p:= allocMem(255);
     formatMessage(FORMAT_MESSAGE_FROM_SYSTEM, nil, errorCode, 0, p, 254, nil);
     result:= trim(p);
     freeMem(p);
end;

function initHawkDI: boolean;
var
  lErrorMode:Integer;
begin
     codecNames[HV_2_4K_CODEC]       := 'LPC-10 2.4 Kbps codec';
     codecNames[HV_4_8K_CODEC]       := 'LPC 4.8 Kbps codec';
     codecNames[HV_13_2K_CODEC]      := 'GSM 13.2 Kbps codec';
     codecNames[HV_32K_CODEC]        := 'Intel/DVI ADPCM 32 Kbps codec';
     codecNames[HV_64K_CODEC]        := 'G.711 u-law 64 Kbps codec';
     codecNames[HV_1_4K_CODEC]       := 'OpenLPC 1.4 Kbps codec';
     codecNames[HV_1_8K_CODEC]       := 'OpenLPC 1.8 Kbps codec';
     codecNames[HV_4_5K_CODEC]       := 'CELP 4.5 Kbps codec';
     codecNames[HV_3_0K_CODEC]       := 'CELP 3.0 Kbps codec';
     codecNames[HV_2_3K_CODEC]       := 'CELP 2.3 Kbps codec';
     codecNames[HV_VBR_2_4K_CODEC]   := 'Variable Bit Rate LPC-10 2.4 Kbps max codec';

     lErrorMode:=0;
     Result:=FALSE;
     try
          lErrorMode:=SetErrorMode($8000);
          dllHandle:=LoadLibrary(HVDI_DLL);
          if dllHandle<>0 then begin
               @hvdiNewEncState:=GetProcAddress(dllHandle,'_hvdiNewEncState@0');
               @hvdiNewDecState:=GetProcAddress(dllHandle,'_hvdiNewDecState@0');
               @hvdiDeleteEncState:=GetProcAddress(dllHandle,'_hvdiDeleteEncState@4');
               @hvdiDeleteDecState:=GetProcAddress(dllHandle,'_hvdiDeleteDecState@4');
               @hvdiEncStateSetCodec:=GetProcAddress(dllHandle,'_hvdiEncStateSetCodec@8');
               @hvdiDecStateGetCodec:=GetProcAddress(dllHandle,'_hvdiDecStateGetCodec@4');
               @hvdiPacketIsVoice:=GetProcAddress(dllHandle,'_hvdiPacketIsVoice@8');
               @hvdiPacketDecode:=GetProcAddress(dllHandle,'_hvdiPacketDecode@24');
               @hvdiPacketEncode:=GetProcAddress(dllHandle,'_hvdiPacketEncode@24');
               @hvdiNewVOX:=GetProcAddress(dllHandle,'_hvdiNewVOX@8');
               @hvdiVOX:=GetProcAddress(dllHandle,'_hvdiVOX@12');
               @hvdiDeleteVOX:=GetProcAddress(dllHandle,'_hvdiDeleteVOX@4');
               @hvdiNewRate:=GetProcAddress(dllHandle,'_hvdiNewRate@8');
               @hvdiRateFlow:=GetProcAddress(dllHandle,'_hvdiRateFlow@20');
               @hvdiDeleteRate:=GetProcAddress(dllHandle,'_hvdiDeleteRate@4');
               @hvdiNewAGC:=GetProcAddress(dllHandle,'_hvdiNewAGC@4');
               @hvdiAGC:=GetProcAddress(dllHandle,'_hvdiAGC@12');
               @hvdiDeleteAGC:=GetProcAddress(dllHandle,'_hvdiDeleteAGC@4');
               @hvdiMix:=GetProcAddress(dllHandle,'_hvdiMix@16');
               @hvdiHint:=GetProcAddress(dllHandle,'_hvdiHint@8');
          end else
               raise Exception.create('Error loading ' + HVDI_DLL + ': ' + getWindowsErrorText);
     finally
          SetErrorMode(lErrorMode);
     end;
end;

// free
procedure FreeHawkDI;
begin
  if dllHandle<>0 then
    FreeLibrary(dllHandle);
  dllHandle:= 0;
end;

function hvdiHighestFramesize: integer;
var
encstate: Phvdi_enc_state;
i, fs: integer;
begin
     encstate:= hvdiNewEncState;
     result:= 0;
     for i:= 1 to 11 do begin
          fs:= hvdiEncStateSetCodec(encstate, i);
          if fs > result then result:= fs;
     end;
     hvdiDeleteEncState(encstate);
end;

{$IFDEF AUTOLOAD}
initialization
     initHawkDI;

finalization

     freeHawkDI;

{$ENDIF}

end.
