{
********************
* unit tilthawkVoice.pas
* Version 1.0
* (c) www.tilt.net
* 31/10/2004
*
* HawkVoice is (c) HawkSoftware (http://www.hawksoft.com)
*
* Classes que facilitam a interface entre o Delphi e a biblioteca
* HawkVoice.
*******************
}
unit tiltHawkVoice;

interface

uses
     hawkVoice, tiltWaveIO, windows, classes, sysUtils, math;

const
     //Tamanho do frame (framsize) caso o hvdiSetCodec retorne "0"
     FRAMESIZE_IF_ZERO = 256;

     //Nmero mximo de buffers usados para "cachear" a sada de som
     NUMBER_BUFFERS = 5;

type
     {: Classe bsica para codificao/decodificao de pacotes do
        HawkVoice. }
     TTiltHawkVoice = class(TObject)
     private
          FEncodeCodec: byte;
          encState: Phvdi_enc_state;
          decState: Phvdi_dec_state;
          FFrameSize: integer;

          lastDecodeCodec: byte;

          procedure setEncodeCodec(val: byte);
     public
          constructor create;
          destructor destroy; override;

          {: Codifica o srcBuffer (de tamanho em bytes srcSize) com o codec escolhido e
             salva o resultado em dstBuffer (que ter dstSize bytes vlidos).
             Um bom "chute" para o tamanho de dstBuffer  o tamanho do srcBuffer.
          }
          procedure encode(srcBuffer: pointer; srcSize: integer; dstBuffer: pointer; var dstSize: integer);

          {: Retorna o tamanho do frame do codec escolhido em bytes. }
          function frameSize: integer;

          {: Decodifica o srcBuffer (de tamanho srcSize em bytes) com o decState
             especificado. O resultado  retornado em dstBuffer, que tem dstSize bytes
             definidos.
          }
          procedure decode(decState: Phvdi_dec_state; srcBuffer: pointer; srcSize: integer; dstBuffer: pointer; var dstSize: integer); overload;

          {: Decodifica o srcBuffer (de tamanho srcSize em bytes) com o decState
             padro. O resultado  retornado em dstBuffer, que tem dstSize bytes
             definidos.
          }
          procedure decode(srcBuffer: pointer; srcSize: integer; dstBuffer: pointer; var dstSize: integer); overload;

          {: Retorna o ltimo codec usado para decodificao. }
          function decodeCodec: byte;

          {: Especifica o codec sendo utilizado para codificao dos pacotes. }
          property EncodeCoded: byte read FEncodeCodec write setEncodeCodec;
     end;

     {: Classe para codificao/decodificao de pacotes de voz retirados
        do microfone (por meio da API WaveIn) e enviados ao dispositivo de
        som (API WaveOut). }
     TTiltWaveHawkVoice = class(TObject)
     private
          FGravando: boolean;
          waveIn: TTiltWaveIn;
          hawk: TTiltHawkVoice;
          FOnBufferCheio: TBufferEvent;
          FNumPacketBuffer: integer;

          tempBuff: pointer;
          tempBuffSize: integer;

          decStates: TStringList;

          selfID: cardinal; //id usado na trasmisso de pacotes

          procedure waveInBufferCheio(sender: TObject; buffer: pointer; size: integer);

          function findOrAddChannel(ID: cardinal): TObject;

          function getEncodeCodec: byte;
          procedure setEncodeCodec(val: byte);
          procedure setGravando(val: boolean);
     public
          constructor Create;
          destructor destroy; override;

          {: Indica  engine de decodificao que um novo pacote de som
             chegou. O buffer especificado ser copiado internamente, portanto
             o chamador pode liberar este espao de memria quando quiser.}
          procedure bufferChegou(buffer: pointer; size: integer);

          {: Especifica o ltimo codec utilizado para decodificao. }
          function decodeCodec: byte;

          {: Especifica se o dispositivo de entrada (microfone) est gravando
             ou no dados. }
          property Gravando: boolean read FGravando write setGravando;

          {: Indica qual o codec sendo utilizado para codificao. }
          property EncodeCoded: byte read getEncodeCodec write setEncodeCodec;

          {: Indica quantos pacotes devem ser armazenados (bufferizados) antes
             que a sua reproduo seja iniciada. }
          property NumPacketBuffer: integer read FNumPacketBuffer write FNumPacketBuffer;

          {: Evento que  chamado toda vez que um buffer de entrada de dados
             (microfone) for preenchido. O buffer retornado foi codificado com
             o codec especificado na propriedade EncodeCodec. Esse buffer deve
             ser copiado antes de utilizado para outros fins. }
          property OnBufferCheio: TBufferEvent read FOnBufferCheio write FOnBufferCheio;
     end;


implementation

type
     TInfoBuffer = record
          buffer: pointer;
          size: integer;
          empty: boolean;

          time: cardinal;
     end;
     PInfoBuffer = ^TInfoBuffer;

     THWFooter = packed record
          senderID: cardinal;
     end;
     PHWFooter = ^THWFooter;

     TOutputChannel = class(TObject)
     private
          inBuffers: TList;
          bufferFilledCount: integer; //nmero de buffers com informao
          procedure waveOutBufferTocado(sender: TObject; buffer: pointer; size: integer);
     public
          tempBuffSize: integer;
          tempBuff: pointer;
          decState: Phvdi_dec_state;
          waveOut: TTiltWaveOut;
          hawk: TTiltHawkVoice;
          waveHawk: TTiltWaveHawkVoice;
          tocando: boolean;

          constructor create;
          destructor Destroy; override;
          procedure bufferChegou(buffer: pointer; size: integer);
     end;

const
     FOOTERSIZE = sizeof(THWFooter);

function customRandom(range: integer): integer;
var
vSeed: integer;
begin
     vSeed:= RandSeed;
     Randomize;
     result:= random(range);
     RandSeed:= vSeed;
end;

//===========================================================================================
//                                      TOutputChannel
//===========================================================================================
constructor TOutputChannel.create;
begin
     inherited create;

     inBuffers:= TList.create;
     bufferFilledCount:= 0;

     decState:= hvdiNewDecState;
     waveOut:= TTiltWaveOut.create;
     waveOut.OnBufferTocado:= waveOutBufferTocado;
     waveOut.open;
end;

destructor TOutputChannel.Destroy;
var
i: integer;
begin
     hvdiDeleteDecState(decState);
     waveOut.Free;
     for i:= 0 to inBuffers.count -1 do dispose(inBuffers[i]);
     inBuffers.free;
     inherited destroy;
end;

procedure TOutputChannel.waveOutBufferTocado(sender: TObject; buffer: pointer;
size: integer);
var
i: integer;
info: PInfoBuffer;
begin
     for i:= 0 to inBuffers.count -1 do begin
          info:= inbuffers[i];
          if info.buffer = buffer then begin
               info.empty:= true;
               dec(bufferFilledCount);
               break;
          end;
     end;
end;

procedure TOutputChannel.bufferChegou(buffer: pointer; size: integer);
var
i: integer;
info: PInfoBuffer;
needBuffer: boolean;
sizeDecoded: integer;
begin
     if bufferFilledCount = 0 then tocando:= false;

     sizeDecoded:= tempBuffSize;
     hawk.decode(decState, buffer, size, tempBuff, sizeDecoded);

     needBuffer:= true;
     info:= nil;
     for i:= 0 to inBuffers.Count -1 do begin
          info:= inBuffers[i];
          if info.empty then begin
               needBuffer:= false;
               break;
          end;
     end;

     if needBuffer then begin
          if inBuffers.count >= waveHawk.NumPacketBuffer + NUMBER_BUFFERS then exit;
          new(info);
          fillchar(info^, sizeof(TInfoBuffer), 0);
          inBuffers.add(info);
     end;

     info.empty:= false;
     inc(bufferFilledCount);
     if info.size < sizeDecoded then begin
          ReallocMem(info.buffer, sizeDecoded);
          info.size:= sizeDecoded;
     end;
     move(tempBuff^, info.buffer^, sizeDecoded);

     info.time:= GetTickCount;

     if tocando then //se j estiver tocando, adiciona diretamente
          waveOut.addBuffer(info.buffer, sizeDecoded)
     else if bufferFilledCount > waveHawk.NumPacketBuffer then begin
          //seno, adiciona apenas se tiver um nmero mnimo de buffers
          //para reproduo
          for i:= 0 to inBuffers.Count -1 do begin
               info:= inBuffers[i];
               if not info.empty then
                    waveOut.addBuffer(info.buffer, info.size);
          end;
          tocando:= true;
     end;
end;

//===========================================================================================
//                                      TTiltHawkVoice
//===========================================================================================
constructor TTiltHawkVoice.create;
begin
     inherited create;

     //criar o encode state
     encstate:= hvdiNewEncState;

     //criar o decode state
     decstate:= hvdiNewDecState;

     EncodeCoded:= HV_GSM_CODEC;
end;

destructor TTiltHawkVoice.destroy;
begin
     inherited destroy;

     //liberar encstate
     hvdiDeleteEncState(encstate);

     //liberar decstate
     hvdiDeleteDecState(decstate);
end;

procedure TTiltHawkVoice.setEncodeCodec(val: byte);
begin
     FEncodeCodec:= val;
     FFrameSize:= hvdiEncStateSetCodec(encstate, val)* 2;
     if FFrameSize = 0 then
          FFrameSize:= FRAMESIZE_IF_ZERO;
end;

function TTiltHawkVoice.decodeCodec: byte;
begin
     result:= lastDecodeCodec;
end;

function TTiltHawkVoice.frameSize: integer;
begin
     result:= FFrameSize;
end;

procedure TTiltHawkVoice.encode(srcBuffer: pointer; srcSize: integer;
dstBuffer: pointer; var dstSize: integer);
var
fs: integer;
srcNbSamples, bufferSampleSize: integer;
begin
     srcNbSamples:= srcSize div 2;

     fs:= framesize div 2;

     //o tamanho do buffer  na realidade o maior mltiplo do
     //frameSize contido no buffer
     bufferSampleSize:= (srcNbSamples div fs) * fs;

     //codificar um frame
     dstSize:= hvdiPacketEncode(dstBuffer, bufferSampleSize,
                                srcBuffer, dstSize, nil, encstate);
end;

procedure TTiltHawkVoice.decode(decState: Phvdi_dec_state; srcBuffer: pointer;
srcSize: integer; dstBuffer: pointer; var dstSize: integer);
var
numberSamples: integer;
begin
     //nmero de samples do buffer (tem q ter no mnimo o tamanho do framesize do codec usado na codificao)
     numberSamples:= dstSize div 2;

     //decodific-lo
     dstSize:= hvdiPacketDecode(srcBuffer, srcSize, dstBuffer,
                                numberSamples, nil, decState)*2;

     lastDecodeCodec:= hvdiDecStateGetCodec(decState);
end;

procedure TTiltHawkVoice.decode(srcBuffer: pointer; srcSize: integer;
dstBuffer: pointer; var dstSize: integer);
begin
     decode(self.decState, srcBuffer, srcSize, dstBuffer, dstSize);
end;

//===========================================================================================
//                                      TTiltWaveHawkDI
//===========================================================================================
constructor TTiltWaveHawkVoice.create;
begin
     inherited create;

     decStates:= TStringList.create;
     decStates.Sort;
     selfID:= customRandom(high(integer)) + customRandom(high(integer));

     tempBuffSize:= max(hvdiHighestFramesize*2, FRAMESIZE_IF_ZERO) + FOOTERSIZE;
     tempBuff:= allocMem(tempBuffSize);

     hawk:= TTiltHawkVoice.create;

     waveIn:= TTiltWaveIn.create;
     waveIn.OnBufferCheio:= waveInBufferCheio;
     waveIn.bufferSize:= hawk.frameSize;

     EncodeCoded:= hawk.EncodeCoded;

     NumPacketBuffer:= 100;
end;

destructor TTiltWaveHawkVoice.destroy;
var
i: integer;
begin
     waveIn.free;
     hawk.Free;
     freeMem(tempBuff);
     for i:= 0 to decStates.Count -1 do decStates.Objects[i].Free;
     decStates.free;
     inherited destroy;
end;

function TTiltWaveHawkVoice.getEncodeCodec: byte;
begin
     result:= hawk.EncodeCoded;
end;

procedure TTiltWaveHawkVoice.setEncodeCodec(val: byte);
begin
     hawk.EncodeCoded:= val;
     waveIn.close;
     waveIn.bufferSize:= hawk.frameSize;

     if Gravando then
          waveIn.open;
end;

procedure TTiltWaveHawkVoice.setGravando(val: boolean);
begin
     FGravando:= val;
     if val then
          waveIn.open
     else
          waveIn.close;
end;

function TTiltWaveHawkVoice.decodeCodec: byte;
begin
     result:= hawk.decodeCodec;
end;

function TTiltWaveHawkVoice.findOrAddChannel(ID: cardinal): TObject;
var
i: integer;
obj: TOutputChannel;
begin
     if decStates.Find(intToStr(id), i) then
          result:= decStates.Objects[i]
     else begin
          obj:= TOutputChannel.create;
          result:= obj;
          obj.tempBuffSize:= tempBuffSize;
          obj.tempBuff:= tempBuff;
          obj.hawk:= hawk;
          obj.waveHawk:= self;
          decStates.Insert(i, intToStr(id));
          decStates.Objects[i]:= obj;
     end;
end;

procedure TTiltWaveHawkVoice.waveInBufferCheio(sender: TObject; buffer: pointer;
size: integer);
var
encodedSize: integer;
footer: THWFooter;
begin
     if not assigned(onBufferCheio) then exit;

     encodedSize:= tempBuffSize;
     hawk.encode(buffer, size, tempBuff, encodedSize);

     //inserir footer
     {$WARNINGS OFF}
     footer.senderID:= selfID;
     move(footer, ptr(cardinal(tempBuff) + encodedSize)^, FOOTERSIZE);
     encodedSize:= encodedSize + FOOTERSIZE;
     {$WARNINGS ON}

     OnBufferCheio(self, tempBuff, encodedSize);
end;

procedure TTiltWaveHawkVoice.bufferChegou(buffer: pointer; size: integer);
var
footer: THWFooter;
channel: TOutputChannel;
begin
     assert(tempBuffSize >= size, 'TempBuffSize < size');

     //retirar footer
     {$WARNINGS OFF}
     move(ptr(cardinal(buffer) + size - FOOTERSIZE)^, footer, FOOTERSIZE);
     size:= size - FOOTERSIZE;
     {$WARNINGS ON}

     channel:= TOutputChannel(findOrAddChannel(footer.senderID));
     channel.bufferChegou(buffer, size);
end;

end.
