{
********************
* unit tiltWaveIO.pas
* Version 1.0
* (c) www.tilt.net
* 31/10/2004
*
* Classes para capturar entrada wave (waveIn) e envio de dados 
* dispositivos de sada de wave (waveOut).
*******************
}
unit tiltWaveIO;

interface

uses
     mmSystem, windows, classes, sysUtils, forms;

type
     TBufferEvent = procedure (sender: TObject; buffer: pointer; size: integer) of object;

     {: Classe de controle de entrada de wave. }
     TTiltWaveIn = class(TObject)
     private
          FBufferSize: integer;
          hWI: HWAVEIN;
          inHdrs: array[0..4] of TWaveHdr;

          thread: TThread;

          FOnBufferCheio: TBufferEvent;

          procedure setBufferSize(val: integer);
     protected
          closing: boolean;

          procedure step;
     public
          constructor create;
          destructor destroy; override;

          {: Abre o dispositivo para gravao. }
          procedure open;

          {: Fecha o dispositivo.}
          procedure close;

          {: Tamanho do buffer de leitura. No pode ser mudado aps o
             dispositivo estar aberto. }
          property bufferSize: integer read FbufferSize write setBufferSize;

          {: Chamado quando um buffer foi preenchido com dados. }
          property OnBufferCheio: TBufferEvent read FOnBufferCheio write FOnBufferCheio;
     end;

     TTiltWaveOut = class(TObject)
     private
          hWO: HWAVEOUT;
          headers: TList;
          FOnBufferTocado: TBufferEvent;
          thread: TThread;
     protected
          closing: boolean;
          procedure waveDone(info: PWaveHdr);
     public
          constructor create;
          destructor destroy; override;

          {: Abre o dispositivo de sada. }
          procedure open;

          {: Fecha o dispositivo. }
          procedure close;

          {: Adiciona um buffer para reproduo. }
          procedure addBuffer(buffer: pointer; size: integer);

          {: Chamado quando um buffer foi tocado. }
          property OnBufferTocado: TBufferEvent read FOnBufferTocado write FOnBufferTocado;
     end;

implementation

type
     TWaveInThread = class(TThread)
     protected
          semaforo: integer;
          procedure doCallStep;
          procedure execute; override;
     public
          wavehdr: PWaveHdr;
          waveIn: TTiltWaveIn;
          close: boolean;
     end;

     TWaveOutThread = class(TThread)
     protected
          procedure doCallDone;
          procedure execute; override;
     public
          waveOut: TTiltWaveOut;
          close: boolean;
          hdr: PWaveHdr;
     end;

//===========================================================================================
//                                     TWaveInThread
//===========================================================================================
procedure TWaveInThread.doCallStep;
begin
     if (not assigned(waveIn)) {or (not assigned(waveHDR))} then exit;
     with waveIn do begin
          //if assigned(onBufferCheio) then
            //   OnBufferCheio(waveIn, waveHdr.lpData, waveHdr.dwBufferLength);
          step;
     end;
end;

procedure TWaveInThread.execute;
begin
     while not close do begin
          //writeln('..before suspend..');
          //inc(semaforo);
          //Suspend;
          //writeln('..after suspend..');
          sleep((waveIn.bufferSize div 2) div 8 - 10);
          Synchronize(doCallStep);
          //writeln('semaforo: ' + intToStr(semaforo));
     end;
end;

//===========================================================================================
//                                     TWaveOutThread
//===========================================================================================
procedure TWaveOutThread.doCallDone;
begin
     if not assigned(waveOut) then exit;
     with waveOut do
          waveDone(hdr);
end;

procedure TWaveOutThread.execute;
begin
     while not close do begin
          Suspend;
          Synchronize(doCallDone);
     end;
end;

//===========================================================================================
//                                      TTiltWaveIn
//===========================================================================================
procedure wiCback(wi: HWAVEIN; uMsg: LongWord;
                  dwInstance,dwParam1,dwParam2: LongWord); stdcall;
//var
  //waveIn: TTiltWaveIn;
  //waveHdr: PWaveHdr;
  //inThread: TWaveInThread;
begin
     if dwInstance = 0 then exit;
     //inThread:= TWaveInThread(dwInstance);
     //inThread.wavehdr:= PWaveHdr(dwParam1);
     //while not inThread.Suspended do inThread.resume;
     //writeln('will resume'); inThread.semaforo:= inThread.semaforo -1;
     //inThread.Resume;
     //if inThread.Suspended then writeln('still suspended');

     {waveIn:= TTiltWaveIn(dwInstance);
     waveHdr:= PWaveHdr(dwParam1);
     case uMsg of
          WIM_DATA: with waveIn do begin
               if assigned(onBufferCheio) then
                    OnBufferCheio(waveIn, waveHdr.lpData, waveHdr.dwBufferLength);
               step;
         end;
     end;}
end;

constructor TTiltWaveIn.create;
begin
     inherited create;
     hWI:= -1;
     bufferSize:= 512;
     thread:= TWaveInThread.Create(false);
     TWaveInThread(thread).waveIn:= self;
     thread.FreeOnTerminate:= true;
     //thread.Priority:= tpTimeCritical;
end;

destructor TTiltWaveIn.destroy;
begin
     close;
     TWaveInThread(thread).close:= true;
     TWaveInThread(thread).semaforo:= TWaveInThread(thread).semaforo - 1;
     thread.Resume;
     inherited destroy;
end;

procedure TTiltWaveIn.setBufferSize(val: integer);
begin
     if hWI <> -1 then raise Exception.Create('Cannot set bufferSize after waveIn was open');
     FBufferSize:= val;
end;

procedure TTiltWaveIn.open;
var
  wfx: TWaveFormatEx;
  i: integer;
begin
  with wfx do begin
    wFormatTag      := WAVE_FORMAT_PCM;
    nChannels       := 1;
    nSamplesPerSec  := 8000;
    wBitsPerSample  := 16;
    nBlockAlign     := nChannels*wBitsPerSample div 8;
    nAvgBytesPerSec := nSamplesPerSec*nBlockAlign;
    cbSize          := 0;
  end;

  close;
  if waveInOpen(@hWI,0,@wfx,Cardinal(@wiCback),cardinal(thread),
                CALLBACK_FUNCTION or WAVE_MAPPED)<>MMSYSERR_NOERROR then begin
    hWI:= -1;
    raise Exception.Create('Cannot open WaveIn Device');
    exit;
  end;

  for i:=0 to High(inHdrs) do begin
    inHdrs[i].lpData := AllocMem(bufferSize);
    inHdrs[i].dwBufferLength := bufferSize;
    inHdrs[i].dwFlags := 0;
    inHdrs[i].dwUser := i;

    if waveInPrepareHeader(hWI,@inHdrs[i],sizeof(inHdrs[i]))<>
       MMSYSERR_NOERROR then begin
      hWI:= -1;
      raise Exception.Create('Cannot prepare WaveHdr record');
      exit;
    end;

    if waveInAddBuffer(hWI,@inHdrs[i],sizeof(inHdrs[i]))<>
       MMSYSERR_NOERROR then begin
      hWI:= -1;
      raise Exception.Create('Cannot add WaveHdr buffer');
      exit;
    end;
  end;

  if waveInStart(hWI)<>MMSYSERR_NOERROR then begin
    hWI:= -1;
    raise Exception.Create('Cannot start WaveIn Device');
    exit;
  end;

  if waveInStart(hWI)<>MMSYSERR_NOERROR then begin
    hWI:= -1;
    raise Exception.Create('Cannot start WaveIn Device');
    exit;
  end;
end;

procedure TTiltWaveIn.close;
begin
  if hWi <> -1 then begin
       closing:= true;
       waveInStop(hWI);
       waveInReset(hWI);
       waveInClose(hWI);
       step;
       closing:= false;
  end;
  hWI:= -1;
end;

procedure TTiltWaveIn.step;
var
  i, err: integer;
begin
  for i:=0 to High(inHdrs) do begin
    if (inHdrs[i].dwFlags and WHDR_DONE)<>0 then begin
      inHdrs[i].dwFlags := 0;

      if hWI = -1 then continue;

      if assigned(onBufferCheio) then
          OnBufferCheio(self, inHdrs[i].lpData, inHdrs[i].dwBufferLength);


      if closing then begin
          waveInUnprepareHeader(hWI, @inHdrs[i], sizeof(inHdrs[i]));
          freeMem(inHdrs[i].lpData);
          continue;
      end;

      err:= waveInPrepareHeader(hWI,@inHdrs[i],sizeof(inHdrs[i]));
      if err <> MMSYSERR_NOERROR then begin
        //raise Exception.Create('Cannot prepare WaveHdr record ' + intToStr(err) + intToStr(hwi));
        sleep(1);
        exit;
      end;

      if waveInAddBuffer(hWI,@inHdrs[i],sizeof(inHdrs[i]))<>
         MMSYSERR_NOERROR then begin
        //raise Exception.Create('Cannot add WaveHdr buffer');
        exit;
      end;
    end;
  end;
end;

//===========================================================================================
//                                      TTiltWaveOut
//===========================================================================================
procedure woCback(wi: HWAVEOUT; uMsg: LongWord;
dwInstance,dwParam1,dwParam2: LongWord); stdcall;
begin
     if dwInstance = 0 then exit;
     with TWaveOutThread(dwInstance) do begin
          hdr:= PWaveHDR(dwParam1);
          Resume;
     end;
end;

constructor TTiltWaveOut.create;
begin
     inherited create;
     thread:= TWaveOutThread.create(false);
     thread.FreeOnTerminate:= true;
     TWaveOutThread(thread).waveOut:= self;
     headers:= TList.create;
     hWO:= -1;
end;

destructor TTiltWaveOut.destroy;
var
i: integer;
begin
     close;
     for i:= 0 to headers.count -1 do dispose(headers[i]);
     headers.free;
     TWaveOutThread(thread).close:= true;
     thread.Resume;
     inherited destroy;
end;

procedure TTiltWaveOut.open;
var
  wfx: TWaveFormatEx;
begin
  with wfx do begin
    wFormatTag      := WAVE_FORMAT_PCM;
    nChannels       := 1;
    nSamplesPerSec  := 8000; //smpRt;
    wBitsPerSample  := 16;
    nBlockAlign     := nChannels*wBitsPerSample div 8;
    nAvgBytesPerSec := nSamplesPerSec*nBlockAlign;
    cbSize          := 0;
  end;

  waveOutClose(hWO);
  if waveOutOpen(@hWO,0,@wfx,cardinal(@woCBack),cardinal(thread),CALLBACK_FUNCTION or WAVE_MAPPED)<>MMSYSERR_NOERROR then begin
    raise Exception.Create('Cannot open WaveOut Device');
  end;
end;

procedure TTiltWaveOut.close;
begin
  if hWO <> -1 then begin
       closing:= true;
       waveOutReset(hWO);
       waveOutClose(hWO);
       closing:= false;
  end;
  hWO:= -1;
end;

procedure TTiltWaveOut.addBuffer(buffer: pointer; size: integer);
var
info: PWaveHdr;
begin
     new(info);
     fillchar(info^, sizeof(WaveHdr), 0);
     info.lpData := buffer;
     info.dwBufferLength := size;
     info.dwFlags := 0;

     if waveOutPrepareHeader(hWO,info,sizeof(TWaveHdr))<>
        MMSYSERR_NOERROR then begin
       dispose(info);
       raise Exception.Create('Cannot prepare WaveHdr record');
       exit;
     end;

     if waveOutWrite(hWO,info,sizeof(TWaveHdr))<> MMSYSERR_NOERROR then begin
          dispose(info);
          raise Exception.Create('Cannot write WaveHdr record');
          exit;
     end;

     headers.Add(info);

end;

procedure TTiltWaveOut.waveDone(info: PWaveHDR);
var
//info: PWaveHdr;
i: integer;
begin
     if headers.Count = 0 then exit;
     if info = nil then exit;

     {if (info.dwFlags and WHDR_DONE) <> 0 then begin
          if assigned(onBufferTocado) then
               OnBufferTocado(self, info.lpData, info.dwBufferLength);
     end;}

     i:= 0;
     while i < headers.count do begin
          info:= headers[i];
          if (info.dwFlags and WHDR_DONE) <> 0 then begin
               if assigned(onBufferTocado) then
                    OnBufferTocado(self, info.lpData, info.dwBufferLength);
               //if not closing then
                    waveOutUnprepareHeader(hWO, info, sizeof(WaveHdr));
               headers.Delete(i);
               dispose(info);
               headers.Capacity:= headers.Count;
          end else inc(i);
     end;
end;

end.
